//***************************************************************************
// Group Linslot / Linux - Slotrace Manager
// File iothread.cc
// Date 28.11.06 - Jrg Wendel
// This code is distributed under the terms and conditions of the
// GNU GENERAL PUBLIC LICENSE. See the file COPYING for details.
//***************************************************************************

//***************************************************************************
// Includes
//***************************************************************************

#include <sys/time.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <QSqlQuery>
#include <QVariant>

#include <iothread.hpp>
#include <common.hpp>

#ifdef WITH_IOWARRIOR
#  include <iow.hpp>
#endif

#ifdef WITH_ARDUINO
#  include <arduino.hpp>
#endif

//***************************************************************************
// Class IoThread
//***************************************************************************
//***************************************************************************
// Object
//***************************************************************************

IoThread::IoThread()
   : QThread()
{
   testMode = no;
   running = no;
   *device = 0;
   command = cNone;
   gcPause = 0;
   active = no;

   gcReplayTimer = new QTimer();

   connect(gcReplayTimer, SIGNAL(timeout()), this, 
           SLOT(onReplayTimer()));

#ifdef WITH_IOWARRIOR
   ioDevice = new Iow;
#elif WITH_ARDUINO
   ioDevice = new Arduino;
#else
   ioDevice = 0;
#endif
}

IoThread::~IoThread()
{
   exit();

   if (ioDevice) delete ioDevice;
}

//***************************************************************************
// Init
//***************************************************************************

int IoThread::init()
{  
   if (!ioDevice)
   {
      tell(eloAlways, "No IO device compiled in, aborting!");
      return fail;
   }

   return success;
}

//***************************************************************************
// Set Device
//***************************************************************************

void IoThread::setDevice(const char* dev)
{
   if (!dev)
      return ;

   if (strcmp(device, dev) != 0)
   {
      strncpy(device, dev, 100); 
      device[100] = 0;
   }
}

//***************************************************************************
// Set Device
//***************************************************************************

void IoThread::changeDevice(const char* dev)
{
   if (!dev)
      return ;

   if (strcmp(device, dev) != 0)
   {
      setDevice(dev);

      close();
      open();
   }
}

//***************************************************************************
// Exit
//***************************************************************************

int IoThread::exit()
{
   // Close device
   
   close();

   return success;
}

//***************************************************************************
// Open
//***************************************************************************

int IoThread::open()
{
   close();

   if (ioDevice->open(device) != success)
   {
      tell(eloAlways, "Failed to open device '%s<N>'", device);
      return fail;
   }

   ioDevice->setWriteTimeout(1000);
   ioDevice->setTimeout(1000);

   return success;
}

//***************************************************************************
// Close
//***************************************************************************

int IoThread::close()
{
   ioDevice->close();
   
   return 0;
}

void IoThread::stopGhostCar()
{
   tell(eloAlways, "Stopping ghost car");

   gcReplayTimer->stop();
   usleep(200);
   gcValues.clear();
   gcValueIndex = 0;
   gcPause = 0;

   ioDevice->stopGhostCar();
}

void IoThread::startGhostCar(char pwmBit, char iBit, int profileId)
{
   GcValue value;
   int count = 0;

   gcValueIndex = 0;

   QSqlQuery q("select VOLT, AMPERE from lap_profiles where PROFILE_ID = '" + 
               QString::number(profileId) + "';");

   while (q.next())
   {
      gcValues.append(value);

      value.volt = (byte)q.value(0).toInt();
      value.ampere = (byte)q.value(1).toInt();
      count++;
   }

   if (count)
   {
      tell(eloAlways, "Starting ghost car");

      ioDevice->startGhostCar(pwmBit, iBit);
      gcPause = 0;
      gcReplayTimer->start(20);
      tell(eloAlways, "added %d values", count);
   }
   else
   {
      tell(eloAlways, "No data for profile %d found", profileId);
   }
}

void IoThread::onReplayTimer()
{
   if (!gcValues.size())
      return;

   if (gcPause > 0)
   {
      gcPause--;
      return;
   }

   if (gcValueIndex >= gcValues.size())
   {
      gcValueIndex = 0;
      tell(eloAlways, "Reset index, starting next lap");
   }

   ioDevice->writeGhostCarValue(gcValues.at(gcValueIndex).volt, 
                                gcValues.at(gcValueIndex).ampere);
   gcValueIndex++;
}

//***************************************************************************
// At Trouble
//***************************************************************************

int IoThread::atTrouble()
{
   const int retryEvery = 10;
   static int lastTry = time(0);

   tell(eloAlways, "Retrying to open in %d seconds", lastTry + retryEvery - time(0));

   if (time(0) < lastTry + retryEvery)
      return ignore;

   if (!ioDevice->isOpen())
   {
      tell(eloAlways, "Retrying to open device");
      lastTry = time(0);

      return open();
   }

   return done;
}

//***************************************************************************
// Look IO
//***************************************************************************

int IoThread::lookIo()
{
   // needed ... ?

   return done;
}

//***************************************************************************
// Control
//***************************************************************************

void IoThread::control()
{
   if (!active)
      return ;

   switch (command)
   {
      case cDigitalIn: 
      {
         DigitalEvent event;
         DigitalInput* input;

         if (ioDevice->getMessage())
         {
            input = (DigitalInput*)ioDevice->getMessage();
            event.value = input->value;
            event.tp = addMs2Tv(ioDevice->getBoardStartTime(), input->time);

            emit onDigitalInput(event);
         }
     
         break;
      }  
      case cAnalogIn: 
      {
         AnalogEvent event;
         AnalogInput* input;

         if (ioDevice->getMessage())
         {
            input = (AnalogInput*)ioDevice->getMessage();
            event.volt = input->volt;
            event.ampere = input->ampere;

            emit onAnalogInput(event);
         }

         break;
      }  
      case cDebug: 
      {
         DebugValue* debug;
         
         if (ioDevice->getMessage())
         {
            debug = (DebugValue*)ioDevice->getMessage();
            
            tell(eloAlways, "<- [%s] (0x%lx)", debug->string, debug->value);
         }

         break;
      }  
      case cGhostCarBufferFull:
      {
         gcPause = 50;  // wait 0.5 seconds
         gcValueIndex--;
         tell(eloDetail, "<- buffer full, waiting");

         break;
      }

      default: break;
   }
}

//***************************************************************************
// Run
//***************************************************************************

void IoThread::run()
{
   int status;

   running = yes;

   // first open io device
   // and flush old input data

   open();
   ioDevice->flush();

// TODO impement test mode again
//    unsigned int* value;
//    usleep(100000);
//    if (testMode)
//    {
//       static unsigned char last = 0x00;
//       last = ~last;
//       *value = last;

//       return success;
//    }

   // loop

   while (running)
   {
      if (!ioDevice->isOpen())
      {
         atTrouble();
         sleep(1);

         continue ;
      }

      if ((status = ioDevice->look(command)) == success)
      {
         control();
         continue;
      }

      usleep(100);
   }
}
