//***************************************************************************
// Group Linslot / Linux - Slotrace Manager
// File linslot.cc
// Date 05.12.06 - Jrg Wendel
// This code is distributed under the terms and conditions of the
// GNU GENERAL PUBLIC LICENSE. See the file COPYING for details.
//***************************************************************************

//***************************************************************************
// Include
//***************************************************************************

#include <sys/time.h>

#include <QMessageBox>
#include <QMetaType>
#include <QSound>
#include <QGraphicsView>
#include <QDialog>
#include <QInputDialog>
#include <QSqlQuery>
#include <QSqlTableModel>
#include <QSplitter>
#include <QHeaderView>

#include <sqlite.hpp>
#include <linslot.hpp>
#include <lapprofile.hpp>

//***************************************************************************
// Object
//***************************************************************************

LinslotWindow::LinslotWindow()
   : QMainWindow()
{
   // first init class privates

   testMode = no;
   theEloquence = eloAlways;
   raceRunning = no;
   countdownStarted = no;
   countdown = 0;
   fastLapTime = 0;
   resourcePath = 0;
   configPath = QString(QDir::homePath() + "/.linslot");
   gcState = gcsOff;
   gcSlot = na;

   QDir d(configPath);

   if (!d.exists())
      d.mkdir(configPath);

   // now setup the user interface

   setupUi(this);
   
   // dialogs

   setupDialog = new SetupDialog(configPath);
   highscoreDialog = new HighscoreDialog();

   connect(setupDialog, SIGNAL(accepted()),
           this, SLOT(onOptionsAccepted()));

   // init

   init();

   // init timer

   timer = new QTimer(this);
   connect(timer, SIGNAL(timeout()), this, SLOT(onTimer()));

   timerElapsed = new QTimer(this);
   connect(timerElapsed, SIGNAL(timeout()), this, SLOT(onElapsedTimer()));

   timerPenalty = new QTimer(this);
   connect(timerPenalty, SIGNAL(timeout()), this, SLOT(onPenaltyTimer()));

   // thread stuff 

   thread = new IoThread();

   // connect thread

   qRegisterMetaType<DigitalEvent>("DigitalEvent");
   qRegisterMetaType<AnalogEvent>("AnalogEvent");

   connect(thread, SIGNAL(onDigitalInput(const DigitalEvent)),
           this, SLOT(onDigitalInput(const DigitalEvent)));

   connect(thread, SIGNAL(onAnalogInput(const AnalogEvent)),
           this, SLOT(onAnalogInput(const AnalogEvent)));

   // apply setup options

   applyOptions();

   initInterface();
}

LinslotWindow::~LinslotWindow()
{
   exit();
   
   free(resourcePath);
   delete highscoreDialog;
   delete setupDialog;
   delete db;
   delete thread;
}

//***************************************************************************
// init/exit
//***************************************************************************

void LinslotWindow::init()
{
   // get setup parameters

   asprintf(&resourcePath, "%s", setupDialog->getResourcePath());

   // widged stuff

   ledsParent = labelLed00->parent();           // !!
   tableWidgetSlot1->setColumnWidth(0, 125);
   tableWidgetSlot1->setColumnWidth(1, 130);
   tableWidgetSlot2->setColumnWidth(0, 125);
   tableWidgetSlot2->setColumnWidth(1, 130);
   tableWidgetSlot1->setSelectionBehavior(QAbstractItemView::SelectRows);
   tableWidgetSlot2->setSelectionBehavior(QAbstractItemView::SelectRows);
   tableWidgetSlot1->setEditTriggers(QAbstractItemView::NoEditTriggers);
   tableWidgetSlot2->setEditTriggers(QAbstractItemView::NoEditTriggers);

   // parse arguments ..

   for (int i = 0; i < QCoreApplication::arguments().size(); i++)
   {
      const char* arg = QCoreApplication::arguments().at(i).toAscii();

      if (arg[0] != '-' || strlen(arg) != 2)  continue;

      switch (arg[1])
      {
         case 't':
         {
            testMode = yes;
            break;
         }
         case 'e':
         {
            i++;
            theEloquence = QCoreApplication::arguments().at(i).toInt();
            tell(eloAlways, "Set eloquence to (%d)", theEloquence);
            break;
         }
      }
   }

   //

   resetWidgets();

   // db

   tell(eloAlways, "open db");

   if (openDb() != success)
   {
      pushButtonSaveToDb->setEnabled(false);
      pushButtonHallOfFame->setEnabled(false);
   }

   tell(eloAlways, "open db done");

   // at last, check sound

   withSound = no;

//    if (!QSound::isAvailable())
//       tell(eloAlways, "No sound available, missing \"Network audio System (NAS)\"");
//    else
//       withSound = yes;
}

void LinslotWindow::exit()
{
   // db

   if (db) db->close();

   storeConfig();

   // stop thread

   tell(eloAlways, "Stopping thread");
   thread->stop();

   if (!thread->wait(5000))
      tell(eloAlways, "Thread would not end!");
   else 
      tell(eloAlways, "Thread ended regularly");
}

//***************************************************************************
// 
//***************************************************************************

int LinslotWindow::initInterface()
{
   // start io thread 

   thread->init();
   thread->setDevice(setupDialog->getUsbDevice());
   thread->start();
   thread->setTestMode(testMode);

   if (testMode)
      return success;

   tell(eloDebug, "Debug: Waiting for io device");

   // TODO, besser via signal anstoen statt hier warten !

   while (!thread->isOpen())
      sleep(1);

   // init 

   thread->initIoSetup(setupDialog->getInputMask(), 
                       setupDialog->getOutputMask());

   usleep(100000);
   thread->activate();

   // switch slot power off

   setSlotPower(psAll, off);
   setLeds(lsAllOff | lsYellow1 | lsYellow2);

   return success;
}

//***************************************************************************
// Store Config
//***************************************************************************

void LinslotWindow::storeConfig()
{
   // save config

   setupDialog->getSettings()->beginGroup("common");
   setupDialog->getSettings()->setValue("driver0", theSlots[0].driver);
   setupDialog->getSettings()->setValue("driver1", theSlots[1].driver);
   setupDialog->getSettings()->endGroup();
}

//***************************************************************************
// Reset Widgets
//***************************************************************************

void LinslotWindow::resetWidgets()
{
   tableWidgetSlot1->setEnabled(false);
   tableWidgetSlot2->setEnabled(false);
   tableWidgetSlot1->clear();
   tableWidgetSlot2->clear();
   tableWidgetSlot1->setEnabled(true);
   tableWidgetSlot2->setEnabled(true);

   QStringList header;
   header << "Zeit" << "km/h";
   tableWidgetSlot1->setHorizontalHeaderLabels(header);
   tableWidgetSlot2->setHorizontalHeaderLabels(header);

   labelInfo->setText("Standby");
   labelFastLap->setText("");

   tableWidgetSlot1->setRowCount(setupDialog->getLapCount());
   tableWidgetSlot2->setRowCount(setupDialog->getLapCount());
   labelLapCounter1->setText(toStr(setupDialog->getLapCount()));
   labelLapCounter2->setText(toStr(setupDialog->getLapCount()));

   if (setupDialog->getFuelingActive())
   {
      progressBarFuelSlot1->setMaximum(QVariant(setupDialog->getFuelMax()).toInt());
      progressBarFuelSlot2->setMaximum(QVariant(setupDialog->getFuelMax()).toInt());
      progressBarFuelSlot1->setValue(QVariant(setupDialog->getFuelMax()).toInt());
      progressBarFuelSlot2->setValue(QVariant(setupDialog->getFuelMax()).toInt());
   }
   else
   {
      progressBarFuelSlot1->setVisible(false);
      progressBarFuelSlot2->setVisible(false);
   }
}

//***************************************************************************
// Get Images For
//***************************************************************************

int LinslotWindow::getImagesFor(int led, const char* &ledOn, const char* &ledOff)
{
   switch (led)
   {
      case lsRed1:
      case lsRed2:
      case lsRed3:
      case lsRed4:
      case lsRed5: 
         ledOn  = "/pixmap/led-red-on.gif";
         ledOff = "/pixmap/led-red-off.gif";
         break;
      case lsGreen:
         ledOn  = "/pixmap/led-green-on.gif";
         ledOff = "/pixmap/led-green-off.gif";
         break;
      case lsYellow1:
      case lsYellow2:
         ledOn  = "/pixmap/led-yellow-on.gif";
         ledOff = "/pixmap/led-yellow-off.gif";
         break;

      default: return -1;
   }

   return 0;
}

//***************************************************************************
// Set Leds
//***************************************************************************

void LinslotWindow::setLeds(int mask)
{
   const char* ledOn  = "";
   const char* ledOff = "";   

   thread->writeBit(outputBitOf(bitLightGreen), mask & lsGreen);
   thread->writeBit(outputBitOf(bitLightYellow1), mask & lsYellow1);
   thread->writeBit(outputBitOf(bitLightYellow2), mask & lsYellow1);   // due to 2 slots 
   thread->writeBit(outputBitOf(bitLightYellow3), mask & lsYellow2);   // due to 2 slots 
   thread->writeBit(outputBitOf(bitLightYellow4), mask & lsYellow2);   // due to 2 slots 
   thread->writeBit(outputBitOf(bitLightRed1), mask & lsRed1);
   thread->writeBit(outputBitOf(bitLightRed2), mask & lsRed2);
   thread->writeBit(outputBitOf(bitLightRed3), mask & lsRed3);
   thread->writeBit(outputBitOf(bitLightRed4), mask & lsRed4);
   thread->writeBit(outputBitOf(bitLightRed5), mask & lsRed5);

   for (int i = 0; i < ledsParent->children().size(); i++)
   {
      QLabel* label = qobject_cast<QLabel*>(ledsParent->children().at(i));

      if (label)
      {
         if (getImagesFor(label->indent(), ledOn, ledOff) != 0)
            continue;

         if (mask & label->indent())
            label->setPixmap(QPixmap(QString(resourcePath) + ledOn));
         else
            label->setPixmap(QPixmap(QString(resourcePath) + ledOff));
      }
   }
}

//***************************************************************************
// Switch Leds
//***************************************************************************

void LinslotWindow::switchLeds(int mask, int state)
{
   const char* ledOn  = "";
   const char* ledOff = "";   

   if (mask & lsGreen)
      thread->writeBit(outputBitOf(bitLightGreen), state);
   if (mask & lsYellow1)
   {
      thread->writeBit(outputBitOf(bitLightYellow1), state);
      thread->writeBit(outputBitOf(bitLightYellow2), state);
   }
   if (mask & lsYellow2)
   {
      thread->writeBit(outputBitOf(bitLightYellow3), state);
      thread->writeBit(outputBitOf(bitLightYellow4), state);
   }
   if (mask & lsRed1)
      thread->writeBit(outputBitOf(bitLightRed1), state);
   if (mask & lsRed2)
      thread->writeBit(outputBitOf(bitLightRed2), state);
   if (mask & lsRed3)
      thread->writeBit(outputBitOf(bitLightRed3), state);
   if (mask & lsRed4)
      thread->writeBit(outputBitOf(bitLightRed4), state);
   if (mask & lsRed5)
      thread->writeBit(outputBitOf(bitLightRed5), state);

   for (int i = 0; i < ledsParent->children().size(); i++)
   {
      QLabel* label = qobject_cast<QLabel*>(ledsParent->children().at(i));

      if (label && mask & label->indent())
      {
         if (getImagesFor(label->indent(), ledOn, ledOff) != 0)
            continue;

         if (state)
            label->setPixmap(QPixmap(QString(resourcePath) + ledOn));
         else
            label->setPixmap(QPixmap(QString(resourcePath) + ledOff));
      }
   }
}

//***************************************************************************
// Set Slot Power
//***************************************************************************

void LinslotWindow::setSlotPower(int slot, int flag)
{
   char buf[4+TB];
   tell(eloDetail, "Set power of slots (%s) to %d", toBinStr(slot, buf, 4), flag);

   if (slot & psSlot1)
   {
      switchLeds(lsYellow1, !flag); 
      thread->writeBit(outputBitOf(bitPowerSlot1), flag);
   }

   if (slot & psSlot2)
   {
      switchLeds(lsYellow2, !flag);
      thread->writeBit(outputBitOf(bitPowerSlot2), flag);
   }

   if (slot & psSlot1 && slot & psSlot2)
      pushButtonPower->setIcon(QIcon(QString(resourcePath) + 
                                     (flag ? "pixmap/led-blue-on.gif" : 
                                      "pixmap/led-blue-off.gif")));
}

//***************************************************************************
// On Options Accepted
//***************************************************************************

void LinslotWindow::onOptionsAccepted()
{
   applyOptions();

   thread->changeDevice(setupDialog->getUsbDevice());
}

void LinslotWindow::applyOptions()
{
   for (int i = 0; i < slotCount; i++)
   {
      theSlots[i].lap = -1;
      theSlots[i].penalty = 0;
      theSlots[i].fuelLevel = setupDialog->getFuelMax();
      *theSlots[i].driver = 0;
      theSlots[i].fueling = no;
      theSlots[i].gcProfile = na;

      theSlots[i].fuelTimer = new QTimer(this);

      connect(theSlots[i].fuelTimer, SIGNAL(timeout()), this, 
              i == 0 ? SLOT(onFuelTimerSlot1()) : SLOT(onFuelTimerSlot2()));
   }

   QListWidgetItem* listItem;
   comboBoxDriver1->clear();
   comboBoxDriver2->clear();

   for (int i = 0; i < setupDialog->listWidgetDriver->count(); i++)
   {
      if ((listItem = setupDialog->listWidgetDriver->item(i)))
      {
         comboBoxDriver1->addItem(listItem->text());
         comboBoxDriver2->addItem(listItem->text());
      }
   }

   QSqlQuery query("select NAME from profiles where ACTIVE = 'true';");
   
   while (query.next())
      comboBoxDriver1->addItem("GC: " + query.value(0).toString());

   setupDialog->getSettings()->beginGroup("common");

   int i;

   if ((i = comboBoxDriver1->findText(setupDialog->getSettings()->value("driver0", "").toString())) >= 0)
      comboBoxDriver1->setCurrentIndex(i);

   if ((i = comboBoxDriver2->findText(setupDialog->getSettings()->value("driver1", "").toString())) >= 0)
      comboBoxDriver2->setCurrentIndex(i);

   setupDialog->getSettings()->endGroup();

   resetWidgets();
}

//***************************************************************************
// On Driver1 Index Changed
//***************************************************************************

void LinslotWindow::on_comboBoxDriver1_currentIndexChanged(QString value)
{
   strncpy(theSlots[0].driver, value.toAscii(), sizeName);

   if (QString(value).indexOf("GC: ") == 0)
   {
      QSqlQuery query("select PROFILE_ID from profiles where ACTIVE = 'true';");

      if (query.next())
         theSlots[0].gcProfile = query.value(0).toInt();
      else
         tell(eloAlways, "Fatal: Profile for '%s' not found", theSlots[0].driver);
   }
}

//***************************************************************************
// On Driver2 Index Changed
//***************************************************************************

void LinslotWindow::on_comboBoxDriver2_currentIndexChanged(QString value)
{
   strncpy(theSlots[1].driver, value.toAscii(), sizeName);
}

//***************************************************************************
// On Penalty Timer
//***************************************************************************

void LinslotWindow::onPenaltyTimer()
{
   int n = 0;

   tell(eloDebug2, "onPenaltyTimer");

   for (int i = 0; i < slotCount; i++)
   {
      if (theSlots[i].penalty > 0)
      {
         theSlots[i].penalty--;

         if (theSlots[i].penalty <= 0)
            setSlotPower(i+1, on);
         else
            n += theSlots[i].penalty;
      }
   }

   if (n <= 0)
      timerPenalty->stop();
}

//***************************************************************************
// On Elapsed Timer
//  - race/trainging 'time limit' reached
//***************************************************************************

void LinslotWindow::onElapsedTimer()
{
   timeval tp;
   int finished = no;

   gettimeofday(&tp, 0);
   
   int usec = elapsed(&raceStart, &tp);

   char* buf;
   asprintf(&buf, "%4.1f", (double)usec/1000000L);
   labelElapsed->setText(buf);
   free(buf);

   if (radioButtonLapRace->isChecked() && setupDialog->getMaxTimeLapRace())
      finished = usec/1000000 >= setupDialog->getMaxTimeLapRace();
   else if (radioButtonTraining->isChecked() && setupDialog->getMaxTimeTraining())
      finished = usec/1000000 >= setupDialog->getMaxTimeTraining();
   
   if (finished)
   {
      int s = theSlots[0].lap == theSlots[1].lap ? -1 : 
         theSlots[0].lap > theSlots[1].lap ? 0 : 1;

      tell(eloDetail, "Finished after %d seconds of race/training", usec/1000000);
      atFinish(!radioButtonTraining->isChecked() ? s : -1);
   }
}

//***************************************************************************
// Play Sound
//***************************************************************************

void LinslotWindow::playSound(QString file)
{
   if (withSound)
      QSound::play(QString(resourcePath) + file);
}

//***************************************************************************
// On Timer
//***************************************************************************

void LinslotWindow::onTimer()
{
   if (countdown == 5)
   {
      switchLeds(lsPhase5, off);
      switchLeds(lsGreen, on);
      atStart();
   }
   else
   {
      playSound("/sound/beep.wav");

      switch (countdown)
      {
         case 0: switchLeds(lsPhase1, on); break;
         case 1: switchLeds(lsPhase2, on); break;
         case 2: switchLeds(lsPhase3, on); break;
         case 3: switchLeds(lsPhase4, on); break;
         case 4: switchLeds(lsPhase5, on); break;
      }
      
      countdown++;
   }
}

//***************************************************************************
// On Analog Input
//***************************************************************************

void LinslotWindow::onAnalogInput(const AnalogEvent ioEvent)
{
   byte volt = ioEvent.volt;

   if (setupDialog->getGhostcarInvert())
      volt = 255 - volt;

   int u = (int)((double)volt / 255.0 * 100.0);
   int i = (int)((double)ioEvent.ampere / 255.0 * 100.0);

   if (gcState == gcsRecording)
   {
      labelFastLap->setText(QString::number(u) + "%");

      tell(eloAlways, "-> U = %d%%; I = %d%% [%d/%d]", 
           u, i, volt, ioEvent.ampere);
      gcValues.append(volt | (ioEvent.ampere << 8));
   }
}

//***************************************************************************
// On I/O Signal
//***************************************************************************

void LinslotWindow::onDigitalInput(const DigitalEvent ioEvent)
{
   DigitalEvent theEvent = ioEvent;
   unsigned int value;

   // entprellen

   if (!(value = getChanges(theEvent)))
      return ;

   char buf[100+TB];
   tell(eloDebug, "Debug: Changes detected (%s)", toBinStr(value, buf));

   // externer Start/Stop Taster (nicht im test mode) ?
   
   if (isBit(value, bitStartKey) && !testMode)
   {
      if (!countdownStarted)
            atStartCountdown();
      else
         atAbort("Abbruch");
   }

   // process lap signals
   
   if (isBit(value, bitIrSlot1))
      atSlotSignal(0, &theEvent.tp);
   
   if (isBit(value, bitIrSlot2))
      atSlotSignal(1, &theEvent.tp);

   // process fueling signals

   if (isBit(value, bitFuelStartSlot1))
      atFuelSignal(0, &theEvent.tp, yes);

   if (isBit(value, bitFuelStartSlot2))
      atFuelSignal(1, &theEvent.tp, yes);

   if (isBit(value, bitFuelEndSlot1))
      atFuelSignal(0, &theEvent.tp, no);

   if (isBit(value, bitFuelEndSlot2))
      atFuelSignal(1, &theEvent.tp, no);
}

//***************************************************************************
// On Button Power
//***************************************************************************

void LinslotWindow::on_pushButtonPower_clicked()
{ 
   setSlotPower(psAll, !thread->readOutBit(outputBitOf(bitPowerSlot1)));
}

//***************************************************************************
// On Button Start
//***************************************************************************

void LinslotWindow::on_pushButtonStartRace_clicked()
{
   if (!countdownStarted)
      atStartCountdown();
   else
      atAbort("Abbruch");
}

//***************************************************************************
// On Button Options
//***************************************************************************

void LinslotWindow::on_pushButtonOptions_clicked()
{
   storeConfig();
   setupDialog->show();
}

//***************************************************************************
// On Button Hall Of Fame
//***************************************************************************

void LinslotWindow::on_pushButtonHallOfFame_clicked()
{
   if (!db)
      return ;

   highscoreDialog->setDb(db);
   highscoreDialog->fill();
   highscoreDialog->show();
}

//***************************************************************************
// On Button Save To Db
//***************************************************************************

void LinslotWindow::on_pushButtonSaveToDb_clicked()
{
   saveRace();
}

//***************************************************************************
// At Start
//***************************************************************************

void LinslotWindow::atStartCountdown()
{
   countdownStarted = yes;
   countdown = 0;

   // init slot data

   for (int i = 0; i < slotCount; i++)
   {
      theSlots[i].penalty = 0;
      theSlots[i].lap = -1;
      theSlots[i].fuelLevel = setupDialog->getFuelMax();
   }

   // power on

   setSlotPower(psAll, on);

   // set/reset widget stuff

   radioButtonTraining->setEnabled(radioButtonTraining->isChecked());
   radioButtonLapRace->setEnabled(radioButtonLapRace->isChecked());
   pushButtonOptions->setEnabled(false);

   resetWidgets();
   labelInfo->setText("Countdown");
   pushButtonStartRace->setText("&Abbrechen");
   labelElapsed->setText("0.0");
   timer->start(1000);
}

//***************************************************************************
// At Start
//***************************************************************************

void LinslotWindow::atStart()
{ 
   // init

   countdown = 0;
   fastLapTime = 0;
   raceRunning = yes;
   gettimeofday(&raceStart, 0);
   timer->stop();
   timerElapsed->start(100);          // 0.1 sec
   timerPenalty->start(1000);         // 1.0 sec

   if (QString(theSlots[0].driver).indexOf("GC: ") == 0)
   {
      tell(eloDebug, "Starting ghost car '%s' profile (%d)", 
           theSlots[0].driver, theSlots[0].gcProfile);

      thread->startGhostCar(outputBits[bitPwmOutSlot1].bit, 
                            bitGhostISlot1, theSlots[0].gcProfile);

      gcState = gcsRunning;
   }

   for (int i = 0; i < slotCount; i++)
      theSlots[i].lastSignal = raceStart;

   if (radioButtonLapRace->isChecked())
      labelInfo->setText("Rennen luft");
   else
      labelInfo->setText("Freies Training");

   playSound("/sound/start.wav");
}

//***************************************************************************
// At Finish
//***************************************************************************

void LinslotWindow::atFinish(int slot)
{   
   timeval tp;
   int usec;
   char* buf;

   gettimeofday(&tp, 0);

   if (slot >= 0)
      asprintf(&buf, "! ENDE ! Sieger %s", theSlots[slot].driver);
   else
      asprintf(&buf, "! ENDE !");

   atStop(buf);
   free(buf);

   playSound("/sound/finish.wav");

   usec = elapsed(&raceStart, &tp);

   tell(eloDebug, "Total: %2.2d:%2.2d.%06d",
        usec/1000000/60, usec/1000000%60, usec%1000000);
  
   for (int i = 0; i < slotCount; i++)
      tell(eloDebug, "Bahn %d - %d Runden gefahren", i+1, theSlots[i].lap);
}

//***************************************************************************
// At Abort
//***************************************************************************

void LinslotWindow::atAbort(const char* info)
{
   atStop(info);

   for (int i = 0; i < 5; i++, usleep(100000))
      playSound("/sound/beep.wav");
}

//***************************************************************************
// At Stop
//***************************************************************************

void LinslotWindow::atStop(const char* info)
{
   timeval tp;
   char* buf;
   int usec;

   raceRunning = no;
   countdownStarted = no;
   timer->stop();
   timerElapsed->stop();
   
   // slot power off

   setLeds(lsAllOff | lsYellow1 | lsYellow2);
   setSlotPower(psAll, off);

   if (gcState == gcsRunning)
   {
      thread->stopGhostCar();
      tell(eloDebug, "Stopped chost car");
      gcState = gcsOff;
   }

   // update widgets

   gettimeofday(&tp, 0);
   usec = elapsed(&raceStart, &tp);
   asprintf(&buf, "%4.1f", (double)usec/1000000L);
   labelElapsed->setText(buf);
   free(buf);

   pushButtonStartRace->setText("St&arten");

   labelInfo->setText(info);

   radioButtonTraining->setEnabled(true);
   radioButtonLapRace->setEnabled(true);
   pushButtonOptions->setEnabled(true);
}

//***************************************************************************
// At Jump The Gun
//***************************************************************************
  
void LinslotWindow::atJumpTheGun(int slot)
{
   char* info;

   if (setupDialog->getAbortAtJumpTheGun())
   {
      asprintf(&info, "Abbruch Frhstart '%s'", theSlots[slot].driver);
      atAbort(info);
      free(info);
   }
   else
   {
      asprintf(&info, "Frhstart '%s'", theSlots[slot].driver);
      labelFastLap->setText(info);
      free(info);

      tell(eloAlways, "Jump the gun on slot %d, time penalty of (%d) seconds", 
           slot+1, setupDialog->getPenaltyAtJumpTheGun());

      setSlotPower(slot+1, off);
      theSlots[slot].penalty += setupDialog->getPenaltyAtJumpTheGun();
   }
}

//***************************************************************************
// At No Fuel
//***************************************************************************

void LinslotWindow::atNoFuel(int slot)
{
   tell(eloAlways, "No fuel on slot %d, time penalty of (%d) seconds", 
        slot+1, setupDialog->getFuelPenaltyTime());

   setSlotPower(slot+1, off);
   theSlots[slot].penalty += setupDialog->getFuelPenaltyTime();
   timerPenalty->start(1000);         // 1.0 sec
}

//***************************************************************************
// 
//***************************************************************************

int LinslotWindow::isActiveEdge(int function, int state)
{
   if (state && inputBits[function].mode == teRising)
      return yes;
   else if (!state && inputBits[function].mode == teFalling)
      return yes;

   return no;
}

int LinslotWindow::inputBitOf(int function)
{
   return inputBits[function].bit;
}

int LinslotWindow::functionOfInput(int bit)
{
   for (int i = 0; i < bitInputCount; i++)
      if (inputBits[i].bit == bit)
         return i;

   return na;
}

int LinslotWindow::outputBitOf(int function)
{
   return outputBits[function].bit;
}

//***************************************************************************
// Bounce Check  (entprellen)
//***************************************************************************

unsigned int LinslotWindow::getChanges(DigitalEvent& ioEvent)
{
   static int initialized = no;
   static timeval lastInputTimes[4*bitsPerByte] = { { 0, 0 } };
   static unsigned int lastValue = 0xFFFFFFFF;

   unsigned int mask, b, state, fctMask;
   int i;

   // init 

   if (!initialized)
   {
      initialized = yes;

      for (i = 0; i < 4*bitsPerByte; i++)
      {
         gettimeofday(&lastInputTimes[i], 0);
         lastInputTimes[i].tv_sec -= 10;
      }
   }

   // detect changes

   unsigned int changedFunctions = 0;

   for (int bit = 0; bit < 4*bitsPerByte; bit++)
   {
      mask = 1 << bit;
      state = (ioEvent.value & mask) ? 1 : 0;
      b =  state << bit;

      if (((lastValue & mask) ^ b)
          && (elapsed(&lastInputTimes[bit], &ioEvent.tp) > bounceTime))
      {
         // bit changed -and- last change older than 'bounceTime'

         tell(eloDebug2, "Bit %d toggled to %d", bit, state);

         for (int fct = 0; fct < bitInputCount; fct++)
         {
            fctMask = 1 << fct;

            if (inputBits[fct].bit == bit)
            {
               if (isActiveEdge(fct, state))
               {
                  tell(eloAlways, "Function '%s' (%d) detected", inputFunctionName(fct), fct);
                  changedFunctions |= fctMask;
                  lastInputTimes[bit] = ioEvent.tp;
               }
            }
         }
      }
   }

   lastValue = ioEvent.value;

   return changedFunctions;
}

//***************************************************************************
// Fueling Timer
//***************************************************************************

void LinslotWindow::onFuelTimerSlot1()
{
   atFuelTimer(0);
}

void LinslotWindow::onFuelTimerSlot2()
{
   atFuelTimer(1);
}

void LinslotWindow::atFuelTimer(int slot)
{
   theSlots[slot].fuelTimer->start(100);  // fuel trigger each 0.1 second

   if (!theSlots[slot].fueling)
   {
      theSlots[slot].fueling = yes;
      tell(eloAlways, "Start fueling slot (%d)", slot);
   }
   else
   {
      theSlots[slot].fuelLevel += setupDialog->getFuelingPerSecond() / 10.0;
      tell(eloAlways, "Added %.2f liter fuel for slot (%d)", 
           setupDialog->getFuelingPerSecond() / 10.0, slot);

      if (theSlots[slot].fuelLevel > setupDialog->getFuelMax())
      {
         theSlots[slot].fuelLevel = setupDialog->getFuelMax();
         theSlots[slot].fueling = no;
         theSlots[slot].fuelTimer->stop();
         tell(eloAlways, "Finished fueling slot (%d)", slot);
      }

      if (slot == 0)
         progressBarFuelSlot1->setValue(QVariant(theSlots[slot].fuelLevel).toInt());
      else
         progressBarFuelSlot2->setValue(QVariant(theSlots[slot].fuelLevel).toInt());
   }
}

//***************************************************************************
// At Fuel Signal
//***************************************************************************

void LinslotWindow::atFuelSignal(int slot, const timeval*, int fuelStartSignal)
{
   if (!raceRunning)
      return ;

   tell(eloAlways, "Got '%s' signal for slot %d", 
        fuelStartSignal ? "start fueling" : "stop fueling", slot);

   if (!setupDialog->getFuelingActive())
      return ;

   if (fuelStartSignal)
   {
      theSlots[slot].fuelTimer->start(setupDialog->getFuelingDelay()*1000);
   }
   else
   {
      theSlots[slot].fuelTimer->stop();
      theSlots[slot].fueling = no;
   }
}

//***************************************************************************
// Decrement Fuel
//***************************************************************************

void LinslotWindow::decrementFuel(int slot, unsigned int usec)
{
   if (theSlots[slot].fuelLevel > 0)
   {
      double f = setupDialog->getFuelPerLap();
      int uAverageLap = (int)(setupDialog->getAverageLap() * 1000000.0);
      double diffPercent = labs(uAverageLap - usec) / (uAverageLap / 100.0);

      diffPercent = qMin(diffPercent, 30.0);

      tell(eloDebug3, "Debug: Diff to average lap %2d,%03d seconds => (%.2f%%), uAverageLap (%u), usec (%u)", 
           (uAverageLap-usec)/1000000L, labs((uAverageLap-usec))%1000000L, 
           diffPercent, uAverageLap, usec);

      double korr = 0;

      if (uAverageLap - usec > 0)
         korr = 100 + diffPercent/10.0 * setupDialog->getFuelFactorFastLap();  // faster
      else
         korr = 100 - diffPercent/10.0 * setupDialog->getFuelFactorSlowLap();  // slower

      theSlots[slot].fuelLevel -= f/100 * korr;
      
      tell(eloAlways, "Calc fuel, decrement with (%.2f liter) => (%.2f); korr was (%.2f)", 
           f/100 * korr, theSlots[slot].fuelLevel, korr);

      if (theSlots[slot].fuelLevel <= 0)
      {
         theSlots[slot].fuelLevel = 0;
         atNoFuel(slot);
      }
   }
   else
   {
      atNoFuel(slot);
   }
}

//***************************************************************************
// At Slot Signal
//***************************************************************************

void LinslotWindow::atSlotSignal(int slot, const timeval* tp)
{
   char* tmp;
   int rnd = 0;
   unsigned int usec = 0;
   
   tell(eloDebug, "Debug: Signal slot %d", slot);

   if (slot == 0 && !thread->readOutBit(outputBitOf(bitPowerSlot1)))
      return; 

   if (slot == 1 && !thread->readOutBit(outputBitOf(bitPowerSlot2)))
      return; 

   playSound("/sound/beep.wav");

   // recording ghostcar 

   if (gcState == gcsRecording && slot == gcSlot)
   {
      tell(eloAlways, "GC recording finished, got (%d) values", gcValues.size());
      gcState = gcsOff;
      gcSlot = na;
      thread->recordGhostCar(na, na);
      setSlotPower(psAll, off);
      storeGcRecording(&gcValues);
      labelInfo->setText("Standby");
      labelFastLap->setText("");

      return ;
   }
   else if (gcState == gcsWaitingStart && slot == gcSlot)
   {
      tell(eloAlways, "GC recording started");
      gcState = gcsRecording;
      labelInfo->setText("Ghostcar");
      thread->recordGhostCar(bitGhostUSlot1, bitGhostISlot1);

      return ;
   }

   if (!countdownStarted && !raceRunning)
      return ;

   // check 'jump the gun' (Fruehstart)

   if (!raceRunning && !testMode)
   {
      atJumpTheGun(slot);
      return ;
   }

   // increment lap

   theSlots[slot].lap++;

   // erstes ueberfahren der Startlinie ?

   if (!theSlots[slot].lap)
      return ;

   // calc

   usec = elapsed(&theSlots[slot].lastSignal, tp);

   tell(eloDetail, "%d:%d)  %2d.%03d Sekunden (%02.2f km/h)",
        slot, theSlots[slot].lap,
        usec/1000000L, usec%1000000L,
        kmh(setupDialog->getSlotLength(), usec));

   if (radioButtonTraining->isChecked() && 
       tableWidgetSlot1->rowCount() <= theSlots[slot].lap)
   {
      tableWidgetSlot1->setRowCount(theSlots[slot].lap+10);
      tableWidgetSlot2->setRowCount(tableWidgetSlot1->rowCount());
   }

   asprintf(&tmp, " %2.3f", ((double)usec)/1000000L);
   QTableWidgetItem* itemTime = new QTableWidgetItem(tmp);
   free(tmp);

   asprintf(&tmp, "%02.2f",  kmh(setupDialog->getSlotLength(), usec)*setupDialog->getSpeedFactor());
   QTableWidgetItem* itemKmh = new QTableWidgetItem(tmp);
   free(tmp);

   if (!fastLapTime || fastLapTime > usec)
   {
      fastLapTime = usec;

      asprintf(&tmp, "Schnellste Runde\n%s  - %2.3fs (%02.2fkm/h)", 
               *theSlots[slot].driver ? theSlots[slot].driver : "",
               ((double)usec)/1000000L, 
               kmh(setupDialog->getSlotLength(), usec)*setupDialog->getSpeedFactor());
      labelFastLap->setText(tmp);
      free(tmp);
   }

   // fueling

   if (setupDialog->getFuelingActive())
      decrementFuel(slot, usec);

   // 

   if (radioButtonLapRace->isChecked())
      rnd = setupDialog->getLapCount()-theSlots[slot].lap;
   else
      rnd = theSlots[slot].lap;

   if (slot == 0)
   {
      labelLapCounter1->setText(toStr(rnd));
      tableWidgetSlot1->setItem(theSlots[slot].lap-1, 0, itemTime);
      tableWidgetSlot1->setItem(theSlots[slot].lap-1, 1, itemKmh);
      tableWidgetSlot1->setCurrentCell(theSlots[slot].lap-1, 0);
      progressBarFuelSlot1->setValue(QVariant(theSlots[slot].fuelLevel).toInt());
   }
   else
   {
      labelLapCounter2->setText(toStr(rnd));
      tableWidgetSlot2->setItem(theSlots[slot].lap-1, 0, itemTime);
      tableWidgetSlot2->setItem(theSlots[slot].lap-1, 1, itemKmh);
      tableWidgetSlot2->setCurrentCell(theSlots[slot].lap-1, 0);
      progressBarFuelSlot2->setValue(QVariant(theSlots[slot].fuelLevel).toInt());
   }

   gettimeofday(&theSlots[slot].lastSignal, 0);

   if (radioButtonLapRace->isChecked() && 
       theSlots[slot].lap >= setupDialog->getLapCount())
   {
      tell(eloDetail, "Race finished after %d laps of slot %d", 
           theSlots[slot].lap, slot);
      atFinish(slot);
   }
   else if (radioButtonTraining->isChecked() && 
            theSlots[slot].lap >= setupDialog->getLapLimit())
   {
      tell(eloDetail, "Training finished after %d laps of slot %d", 
           theSlots[slot].lap, slot);
      atFinish(na);
   }
}

//***************************************************************************
// Open Db
//***************************************************************************

int LinslotWindow::openDb()
{
   char* dbPath;
   int status;

   // db path

   asprintf(&dbPath, "%s/%s", configPath.toAscii().constData(), 
            setupDialog->getDatabaseName());


   // Open database via QT

   QSqlDatabase _db;
   
   _db = QSqlDatabase::addDatabase("QSQLITE");
   _db.setDatabaseName(dbPath);
     
   if (!_db.open())
   {
      tell(eloAlways, "opening database failed :(");

      if (!_db.isDriverAvailable("QSQLITE"))
         QMessageBox::information(0, "Database", "QSQLITE is not available");
   }

   // the old way, used for highscores

   db = new SqliteDb(dbPath);

   if ((status = db->open()) != success)
   {
      tell(eloAlways, "Error: Open database '%s' failed. "  \
           "Result was (%d)", dbPath, status);
   }

   else
   {
      if ((status = db->execute("select * from races;")) != success)
      {
         if (QMessageBox::question(this, "Warnung", "Tabellen nicht gefunden, anlegen?", 
                                   QMessageBox::Yes, QMessageBox::No) == QMessageBox::Yes)
         {
            createDb();
            
            if ((status = db->execute("select * from races;")) != success)
               QMessageBox::warning(this, "Fehler", "Tabellen konnten nicht angelegt werden!");
         }
      }
   }
   
   if (status != success)
   {
      db->close();
      delete db;
      db = 0;
   }

   return status;
}

//***************************************************************************
// Create Db
//***************************************************************************

int LinslotWindow::createDb()
{
   QSqlQuery* query;

   tell(eloDebug, "Creating tables");

   query = new QSqlQuery("CREATE TABLE races ("                          \
                         "RACE_ID INTEGER PRIMARY KEY AUTOINCREMENT, "    \
                         "DRIVER1 INTEGER, "                             \
                         "DRIVER2 INTEGER, "                             \
                         "DATE DATETIME, "                               \
                         "LAPS INTEGER, "                                \
                         "LAP_LENGTH REAL, "                             \
                         "COURSE TEXT"                                   \
                         ");");
   
   delete query;

   query = new QSqlQuery("CREATE TABLE laps ("                         \
                         "LAP_ID INTEGER PRIMARY KEY AUTOINCREMENT, "  \
                         "RACE_ID INTEGER, "                           \
                         "DRIVER_NR INTEGER, "                         \
                         "LAP_NR INTEGER, "                            \
                         "PROFILE_ID INTEGER, "                        \
                         "LAP_TIME REAL"                               \
                         ");");
   
   delete query;
   
   query = new QSqlQuery("CREATE TABLE drivers ("                       \
                         "DRIVER_ID INTEGER PRIMARY KEY AUTOINCREMENT, " \
                         "NAME TEXT"                                    \
                         ");");
   
   delete query;
   
   query = new QSqlQuery("CREATE TABLE courses ("                       \
                         "COURSE_ID INTEGER PRIMARY KEY AUTOINCREMENT, " \
                         "LENGTH REAL, "                                \
                         "NAME TEXT"                                    \
                         ");");
   

   delete query;

   query = new QSqlQuery("CREATE TABLE profiles ("                      \
                         "PROFILE_ID INTEGER PRIMARY KEY AUTOINCREMENT, " \
                         "NAME TEXT, "                                  \
                         "ACTIVE BOOLEAN, "                             \
                         "COURSE TEXT, "                                \
                         "COLOR TEXT, "                                 \
                         "LAP_LENGTH REAL"                              \
                         ");");

   delete query;

   query = new QSqlQuery("CREATE TABLE lap_profiles ("                  \
                         "LAP_PROFILE_ID INTEGER PRIMARY KEY AUTOINCREMENT, " \
                         "PROFILE_ID INTEGER, "                         \
                         "SEQUENCE INTEGER, "                           \
                         "VOLT INTEGER, "                               \
                         "AMPERE INTEGER"                               \
                         ");");

   delete query;

   return 0;
}


//***************************************************************************
// Show Database
//***************************************************************************

int LinslotWindow::showDb()
{
   db->execute("SELECT r.DRIVER1, r.DRIVER2, r.DATE, l.LAP_TIME from races as r join laps as l where r.RACE_ID = 1 and l.RACE_ID = r.RACE_ID;");
   db->show();

   return 0;
}

//***************************************************************************
// Store Race in Database
//***************************************************************************

int LinslotWindow::saveRace()
{
   sqlite3_stmt* sqlInsertRace;
   sqlite3_stmt* sqlInsertLap;
   int status;
   const char* duration;
   int driver1, driver2;
   int courseId;

   if (!*theSlots[0].driver || !*theSlots[1].driver)
   {
      QMessageBox::critical(this, "Fehler", "Kein Fahrer gewhlt!");
      return 0;
   }

   driver1 = getDriverId(theSlots[0].driver);
   driver2 = getDriverId(theSlots[1].driver);
   courseId = getCourseId();

   db->prepare("INSERT INTO races(DRIVER1,DRIVER2,DATE,LAPS,LAP_LENGTH,COURSE) "\
               "VALUES(?,?,datetime(?, 'unixepoch', 'utc'),?,?,?);", 
               sqlInsertRace);

   db->prepare("INSERT INTO laps(RACE_ID,DRIVER_NR,LAP_NR,LAP_TIME) "\
               "VALUES(?,?,?,?);", 
               sqlInsertLap);

   db->execute("BEGIN;");

   db->reset(sqlInsertRace);
   
   db->bindInt(sqlInsertRace,    1, driver1);
   db->bindInt(sqlInsertRace,    2, driver2);
   db->bindInt(sqlInsertRace,    3, time(0));
   db->bindInt(sqlInsertRace,    4, setupDialog->getLapCount());
   db->bindDouble(sqlInsertRace, 5, setupDialog->getSlotLength());
   db->bindText(sqlInsertRace,   6, setupDialog->getCourseName());
   
   status = db->step(sqlInsertRace);
   
   if (status != 0)
      tell(eloAlways, "sqlite3_step(INSERT INTO races): '%s' (%d)",
           db->lastError(), status);
   
   int raceId = db->getInsertRowId();  // buggy ... ??

   for (int l = 0; l < tableWidgetSlot1->rowCount(); l++)
   {
      db->reset(sqlInsertLap);

      duration = 0;

      if (tableWidgetSlot1->item(l, 1))
         duration = tableWidgetSlot1->item(l, 0)->text().toAscii();

      db->bindInt(sqlInsertLap, 1, raceId);       // RACE_ID
      db->bindInt(sqlInsertLap, 2, driver1);      // DRIVER_NR
      db->bindInt(sqlInsertLap, 3, l+1);          // LAP_NR

      if (duration)
         db->bindText(sqlInsertLap, 4, duration); // LAP_TIME
      else
         db->bindNull(sqlInsertLap, 4);           // LAP_TIME
      
      status = db->step(sqlInsertLap);
      
      if (status != 0)
         tell(eloAlways, "sqlite3_step(INSERT INTO laps): %s (%d)",
              db->lastError(), status);
   }

   for (int l = 0; l < tableWidgetSlot2->rowCount(); l++)
   {
      db->reset(sqlInsertLap);
      
      duration = 0;

      if (tableWidgetSlot2->item(l, 1))
         duration = tableWidgetSlot2->item(l, 0)->text().toAscii();

      db->bindInt(sqlInsertLap, 1, raceId);       // RACE_ID
      db->bindInt(sqlInsertLap, 2, driver2);      // DRIVER_NR
      db->bindInt(sqlInsertLap, 3, l+1);          // LAP_NR
      
      if (duration)
         db->bindText(sqlInsertLap, 4, duration); // LAP_TIME
      else
         db->bindNull(sqlInsertLap, 4);           // LAP_TIME

      status = db->step(sqlInsertLap);
      
      if (status != 0)
         tell(eloAlways, "sqlite3_step(INSERT INTO laps): %s (%d)", 
              db->lastError(), status);
   }
   
   db->execute("COMMIT;");

   db->finalize(sqlInsertRace);
   db->finalize(sqlInsertLap);  

   return 0;
}

//***************************************************************************
// Get Driver Id
//***************************************************************************

int LinslotWindow::getDriverId(const char* driver)
{
   sqlite3_stmt* sqlSelectDriver;
   sqlite3_stmt* sqlInsertDriver;
   char* sql;
   int status;
   int id;

   // get driver ID's

   asprintf(&sql, "SELECT * from drivers where NAME=?;");
   db->prepare(sql, sqlSelectDriver);
   db->bindText(sqlSelectDriver, 1, driver);

   status = db->steps(sqlSelectDriver);

   db->reset(sqlSelectDriver);
   db->finalize(sqlSelectDriver);
   free(sql);  

   if (status == 0)
      id = db->getIntValueOf("DRIVER_ID");
   else
   {
      asprintf(&sql, "INSERT INTO drivers(NAME) values(?);");
      db->prepare(sql, sqlInsertDriver);
      db->execute("BEGIN;");
      db->bindText(sqlInsertDriver, 1, driver);
      db->step(sqlInsertDriver);
      db->execute("COMMIT;");
      id = db->getInsertRowId();
      db->finalize(sqlInsertDriver);
      free(sql);
   }

   return id;
}

//***************************************************************************
// Get Driver Id
//***************************************************************************

int LinslotWindow::getCourseId()
{
   sqlite3_stmt* sqlSelect;
   sqlite3_stmt* sqlInsert;
   char* sql;
   int status;
   int id;

   // get driver ID's

   asprintf(&sql, "SELECT * from cources where NAME=?;");
   db->prepare(sql, sqlSelect);
   db->bindText(sqlSelect, 1, setupDialog->getCourseName());

   status = db->steps(sqlSelect);

   db->reset(sqlSelect);
   db->finalize(sqlSelect);
   free(sql);  

   if (status == 0)
      id = db->getIntValueOf("COURSE_ID");
   else
   {
      asprintf(&sql, "INSERT INTO courses(NAME,LENGTH) values(?,?);");
      db->prepare(sql, sqlInsert);
      db->execute("BEGIN;");
      db->bindText(sqlInsert, 1, setupDialog->getCourseName());
      db->bindDouble(sqlInsert, 2, setupDialog->getSlotLength());
      db->step(sqlInsert);
      db->execute("COMMIT;");
      id = db->getInsertRowId();
      db->finalize(sqlInsert);
      free(sql);
   }

   return id;
}

//***************************************************************************
// 
//***************************************************************************

void LinslotWindow::on_toolButtonRecordGhostCar_clicked()
{
   if (gcState == gcsOff)
   {
      tell(eloAlways, "GC waiting for start signal");
      labelInfo->setText("Ghostcar");
      labelFastLap->setText("bereit");
      gcState = gcsWaitingStart;
      gcSlot = 0;
      gcValues.clear();
      setSlotPower(psAll, on);
   }
   else
   {
      thread->recordGhostCar(na, na);
      labelFastLap->setText("abbruch");
      tell(eloAlways, "GC aborted");
      gcState = gcsOff;
      gcSlot = na;
   }
}

//***************************************************************************
// Store Ghostcar Data
//***************************************************************************

void LinslotWindow::storeGcRecording(QList<unsigned short>* values)
{
   QSqlQuery* insert = new QSqlQuery;
   bool ok = true;
   QString name = "";
   QString hint = "";
   int profileId;

   while (ok)
   {
      name = QInputDialog::getText(this, "Speichern ?",
                                   hint + "Name: ", QLineEdit::Normal,
                                   "", &ok);

      if (ok && !name.isEmpty())
      {
         // name schon vergeben ... ?

         QSqlQuery query("select NAME from profiles " 
                         "where NAME = '" + name + "';");

         if (!query.next())
            break ;

         hint = "Name bereits vergeben. ";
      }
   }

   if (ok && !name.isEmpty())
   {
      // create profile record

      insert->prepare("INSERT INTO profiles("
                      "NAME, ACTIVE, COURSE, LAP_LENGTH) "
                      "VALUES(:name, :active, :course, :length);");

      insert->bindValue(":name", name);
      insert->bindValue(":active", true);
      insert->bindValue(":course", setupDialog->getCourseName());
      insert->bindValue(":length", setupDialog->getSlotLength());
      insert->exec();

      // get id of new profile 

      QSqlQuery query("select PROFILE_ID from profiles " 
                      "where NAME = '" + name + "';");

      if (!query.next())
      {
         tell(eloAlways, "Fatal: Can't find new created profile");
         return ;
      }

      profileId = query.value(0).toInt();

      // create lap profile records

      insert->prepare("INSERT INTO lap_profiles("
                      "PROFILE_ID, SEQUENCE, VOLT, AMPERE) " 
                      "VALUES(:prid, :seq, :volt, :ampere);");

      for (int i = 0; i < values->size(); i++)
      {
         insert->bindValue(":prid", profileId);
         insert->bindValue(":seq", i);
         insert->bindValue(":volt", values->at(i) & 0x00FF);
         insert->bindValue(":ampere", values->at(i) >> 8);

         if (!insert->exec())
         {
            tell(eloAlways,  "Daten konnten nicht gespeichert werden (%d)!", i);
            break ;
         }
      }
   }

   applyOptions();

   delete insert;
}

//***************************************************************************
// 
//***************************************************************************

void LinslotWindow::on_toolButtonTest_clicked()
{
   QDialog* profileDialog = new QDialog(this);

   RenderArea* renderArea = new RenderArea(this);
   QGridLayout* mainLayout = new QGridLayout;
   QToolButton* scaleUp = new QToolButton(this);
   QToolButton* scaleDown = new QToolButton(this);
   QToolButton* scrollLeft = new QToolButton(this);
   QToolButton* scrollRight = new QToolButton(this);
   QToolButton* edit = new QToolButton(this);
   QToolButton* remove = new QToolButton(this);
   QCheckBox* checkAmpere = new QCheckBox(this);
   QCheckBox* checkVolt = new QCheckBox(this);
   QCheckBox* checkPower = new QCheckBox(this);

   QTableView* tableView = new QTableView(this);
   QSqlTableModel* model = new QSqlTableModel(this);
   QSplitter* splitter = new QSplitter(this);

   model->setTable("PROFILES");
   model->removeColumn(model->fieldIndex("COLOR"));
   model->removeColumn(model->fieldIndex("ACTIVE"));
   model->setHeaderData(model->fieldIndex("PROFILE_ID"), Qt::Horizontal, "Id");
   model->setHeaderData(model->fieldIndex("NAME"), Qt::Horizontal, "Name");
   model->setHeaderData(model->fieldIndex("COURSE"), Qt::Horizontal, "Strecke");
   model->setHeaderData(model->fieldIndex("LAP_LENGTH"), Qt::Horizontal, "Lnge");

   model->select();
   tableView->setModel(model);
   tableView->setSelectionBehavior(QAbstractItemView::SelectRows);
   tableView->setEditTriggers(QAbstractItemView::NoEditTriggers);
   tableView->setMaximumSize(400, 9999999);
   tableView->horizontalHeader()->setResizeMode(0, QHeaderView::Fixed);
   tableView->horizontalHeader()->setResizeMode(1, QHeaderView::Stretch);
   tableView->horizontalHeader()->setResizeMode(2, QHeaderView::Fixed);
   tableView->horizontalHeader()->setResizeMode(3, QHeaderView::Fixed);
   profileDialog->setWindowTitle("Rundenprofile");

   scaleUp->setAutoRepeat(true);
   scaleUp->setText("+");
   scaleDown->setAutoRepeat(true);
   scaleDown->setText("-");
   scrollLeft->setAutoRepeat(true);
   scrollLeft->setText("<");
   scrollRight->setAutoRepeat(true);
   scrollRight->setText(">");
   edit->setText("bearbeiten");
   remove->setText("lschen");
   checkVolt->setText("Spannung");
   checkAmpere->setText("Strom");
   checkPower->setText("Leistung");
   checkVolt->setChecked(yes);
   checkAmpere->setChecked(yes);
   checkPower->setChecked(yes);

   splitter->addWidget(renderArea);
   splitter->addWidget(tableView);

   QList<int> sizes;
   sizes << 750 << 250;
   splitter->setSizes(sizes);

   mainLayout->addWidget(splitter, 0, 0, 1, 9);

   mainLayout->addWidget(scaleUp, 1, 0);
   mainLayout->addWidget(scaleDown, 1, 1);
   mainLayout->addWidget(scrollLeft, 1, 2);
   mainLayout->addWidget(scrollRight, 1, 3);
   mainLayout->addWidget(checkVolt, 1, 4);
   mainLayout->addWidget(checkAmpere, 1, 5);
   mainLayout->addWidget(checkPower, 1, 6);
   mainLayout->addWidget(remove, 1, 7, Qt::AlignRight);
   mainLayout->addWidget(edit, 1, 8, Qt::AlignRight);

   connect(scaleUp, SIGNAL(clicked(bool)), renderArea, SLOT(scaleUpClicked(bool)));
   connect(scaleDown, SIGNAL(clicked(bool)), renderArea, SLOT(scaleDownClicked(bool)));
   connect(scrollLeft, SIGNAL(clicked(bool)), renderArea, SLOT(scrollLeftClicked(bool)));
   connect(scrollRight, SIGNAL(clicked(bool)), renderArea, SLOT(scrollRightClicked(bool)));
   connect(tableView, SIGNAL(doubleClicked(const QModelIndex&)), renderArea, SLOT(doubleClicked(const QModelIndex&)));
   connect(edit, SIGNAL(clicked(bool)), renderArea, SLOT(editClicked(bool)));
   connect(remove, SIGNAL(clicked(bool)), renderArea, SLOT(removeClicked(bool)));
   connect(checkVolt, SIGNAL(stateChanged(int)), renderArea, SLOT(checkVolt(int)));
   connect(checkAmpere, SIGNAL(stateChanged(int)), renderArea, SLOT(checkAmpere(int)));
   connect(checkPower, SIGNAL(stateChanged(int)), renderArea, SLOT(checkPower(int)));

   profileDialog->setLayout(mainLayout);
   profileDialog->resize(1000, 600);

   renderArea->setTableView(tableView);
   renderArea->setGcScale(thread->getGcScale());
   tableView->resizeColumnsToContents();
   tableView->resizeRowsToContents();

   profileDialog->exec();
   applyOptions();

   delete profileDialog;
}
