//***************************************************************************
// Group Linslot / Linux - Slotrace Manager
// File linslot.hpp
// Date 05.12.06 - Jrg Wendel
// This code is distributed under the terms and conditions of the
// GNU GENERAL PUBLIC LICENSE. See the file COPYING for details.
//***************************************************************************

#ifndef _LINSLOT_H_
#define _LINSLOT_H_

//***************************************************************************
// Includes
//***************************************************************************

#include <QMainWindow>
#include <QThread>
#include <QReadWriteLock>
#include <QTimer>
#include <QDir>

#include <ui_linslot.h>

#include <highscore.hpp>
#include <setup.hpp>
#include <common.hpp>
#include <iothread.hpp>
#include <sqlite.hpp>

//***************************************************************************
// Class LinslotWindow
//***************************************************************************

class LinslotWindow : public QMainWindow, private Ui::LinslotWindow, public SlotService
{
      Q_OBJECT

   public:

      // declarations

      enum Misc
      {
         sizeName       = 50,
         bitsPerByte    = 8,

         bounceTime     = 30000,    // Seconds (0.03 sec)
         slotCount      = 2,

         // the arduino analog io (0-5)

         bitGhostUSlot1   = 0,   // analog in for U of ghost car slot 1
         bitGhostUSlot3   = 1,   // analog in for U of ghost car slot 3
         bitGhostISlot1   = 2,   // analog in for I of ghost car slot 1
         bitGhostISlot3   = 3,   // analog in for I of ghost car slot 3
         bitAnalogInFree0 = 4,   // free
         bitAnalogInFree1 = 5    // free
      };

      enum PowerState
      {
         psSlot1 = 1,
         psSlot2 = 2,
         psSlot3 = 4,
         psSlot4 = 8,

         psAll = psSlot1 | psSlot2 | psSlot3 | psSlot4
      };

      enum LedState
      {
         lsAllOff     = 0,

         lsGreen      = 1,
         lsRed1       = 2,
         lsRed2       = 4,
         lsRed3       = 8,
         lsRed4       = 16,
         lsRed5       = 32,
         lsYellow1    = 64,
         lsYellow2    = 128,

         lsPhase1     = lsRed1,
         lsPhase2     = lsPhase1 | lsRed2,
         lsPhase3     = lsPhase2 | lsRed3,
         lsPhase4     = lsPhase3 | lsRed4,
         lsPhase5     = lsPhase4 | lsRed5,
         lsPhase6     = lsGreen,
         
         lsSlotPower1 = lsYellow1,
         lsSlotPower2 = lsYellow2,
         lsSlotPower  = lsSlotPower1 | lsSlotPower2,

         lsAllOn      = lsPhase5 | lsGreen | lsSlotPower
      };

      enum GhostCarState
      {
         gcsOff,
         gcsWaitingStart,
         gcsRecording,
         gcsRunning
      };

      struct Slot
      {
         timeval lastSignal;
         int lap;
         char driver[sizeName+TB];
         int penalty;
         double fuelLevel;            // fuel tank level
         int fueling;
         QTimer* fuelTimer;
         int gcProfile;
      };

      // object

      LinslotWindow();
      ~LinslotWindow();

      void init();
      void exit();
      int initInterface();

   protected:

      // functions

      void atSlotSignal(int slot, const timeval* tp);
      void atFuelSignal(int slot, const timeval* tp, int fuelStartSignal);
      void atStartCountdown();
      void atStart();
      void atFinish(int slot);
      void atAbort(const char* info);
      void atJumpTheGun(int slot);
      void atNoFuel(int slot);
      void atStop(const char* info);

      int getImagesFor(int led, const char* &ledOn, const char* &ledOff);
      void setLeds(int mask);
      void switchLeds(int mask, int state);

      void setSlotPower(int slot, int flag);
      unsigned int getChanges(DigitalEvent &ioEvent);
      int isActiveEdge(int bit, int value);
      int inputBitOf(int function);
      int functionOfInput(int bit);
      int outputBitOf(int function);

      void resetWidgets();
      void applyOptions();
      void storeConfig();
      void playSound(QString file);
      void atFuelTimer(int slot);
      void decrementFuel(int slot, unsigned int usec);
      void storeGcRecording(QList<unsigned short>* values);

      // db stuff

      int openDb();
      int createDb();
      int saveRace();
      int showDb();
      int getDriverId(const char* driver);
      int getCourseId();

      // data

      int testMode;
      IoThread* thread;
      QTimer* timer;
      QTimer* timerElapsed;
      QTimer* timerPenalty;
      Slot theSlots[slotCount];
      timeval raceStart;
      int raceRunning;
      int countdownStarted;
      int countdown;
      QObject* ledsParent;
      double fastLapTime;
      int withSound;

      SqliteDb* db;

      char* resourcePath;
      QString configPath;

      QList<unsigned short> gcValues;
      int gcState;
      int gcSlot;

      // dialogs
      
      SetupDialog* setupDialog;
      HighscoreDialog* highscoreDialog;

   private slots:

      void onDigitalInput(const DigitalEvent ioEvent);
      void onAnalogInput(const AnalogEvent ioEvent);

      void on_pushButtonPower_clicked();
      void on_pushButtonSaveToDb_clicked();
      void on_pushButtonHallOfFame_clicked();
      void on_pushButtonStartRace_clicked();
      void on_pushButtonOptions_clicked();
      void on_toolButtonTest_clicked();
      void on_toolButtonRecordGhostCar_clicked();
      void onTimer();
      void onFuelTimerSlot1();
      void onFuelTimerSlot2();
      void onElapsedTimer();
      void onPenaltyTimer();
      void onOptionsAccepted();
      void on_comboBoxDriver1_currentIndexChanged(QString value);
      void on_comboBoxDriver2_currentIndexChanged(QString value);
};

//***************************************************************************
#endif // _LINSLOT_H_
